@@ -22,11 +22,11 @@ Net::CardDAVTalk - A library for talking to CardDAV servers
 
 =head1 VERSION
 
-Version 0.05
+Version 0.06
 
 =cut
 
-our $VERSION = '0.05';
+our $VERSION = '0.06';
 
 
 =head1 SYNOPSIS
@@ -419,6 +419,7 @@ sub GetContactAndProps {
     x('C:addressbook-multiget', $Self->NS(),
       x('D:prop',
         x('D:getetag'),
+        x('D:getcontenttype'),
         x('C:address-data'),
         map { x(join ":", @$_) } @$Props,
       ),
@@ -456,9 +457,27 @@ contacts in scalar context again)
 =cut
 
 sub GetContacts {
-  my ($Self, $Path, $Props, %Args) = @_;
+  my ($Self, $Path, $Props) = @_;
   $Props //= [];
 
+  my $data = $Self->GetContactLinks($Path);
+  my @AllUrls = sort keys %$data;
+
+  my ($Contacts, $Errors, $HRefs) = $Self->GetContactsMulti($Path, \@AllUrls, $Props);
+
+  return wantarray ? ($Contacts, $Errors, $HRefs) : $Contacts;
+}
+
+=head2 $self->GetContactLinks($Path)
+
+Returns a hash of href => etag for every contact URL (type: text/(x-)?vcard)
+inside the collection at \$Path.
+
+=cut
+
+sub GetContactLinks {
+  my ($Self, $Path) = @_;
+
   my $Response = $Self->Request(
     'PROPFIND',
     "$Path/",
@@ -471,51 +490,67 @@ sub GetContacts {
     Depth => '1',
   );
 
-  my @Urls;
+  my %response;
   my $NS_C = $Self->ns('C');
   my $NS_D = $Self->ns('D');
   foreach my $Response (@{$Response->{"{$NS_D}response"} || []}) {
-    my $HRef = $Response->{"{$NS_D}href"}{content};
-    next unless $HRef;
+    my $href = $Response->{"{$NS_D}href"}{content};
+    next unless $href;
     if ($Response->{"{$NS_D}prop"}{"{$NS_D}getcontenttype"}) {
       my $type = $Response->{"{$NS_D}prop"}{"{$NS_D}getcontenttype"}{content} || '';
       next unless $type =~ m{text/(x-)?vcard};
     }
-    push @Urls, $HRef;
+    my $etag = $Response->{"{$NS_D}prop"}{"{$NS_D}getetag"}{content} || '';
+    $response{$href} = $etag;
   }
 
-  my (@Contacts, @Errors);
+  return \%response;
+}
 
-  if (@Urls) {
-    my $Response = $Self->Request(
-      'REPORT',
-      "$Path/",
-      x('C:addressbook-multiget', $Self->NS(),
-        x('D:prop',
-          x('D:getetag'),
-          x('C:address-data'),
-          map { x(join ":", @$_) } @$Props,
-        ),
-        map { x('D:href', $_) } @Urls,
+=head2 $self->GetContactsMulti($Path, $Urls, $Props)
+
+Does an addressbook-multiget on the \$Path for all the URLs in \$Urls
+also fetching \$Props on top of the address-data and getetag.
+
+=cut
+
+sub GetContactsMulti {
+  my ($Self, $Path, $Urls, $Props) = @_;
+  my (@Contacts, @Errors, %ETags);
+
+  my $Response = $Self->Request(
+    'REPORT',
+    "$Path/",
+    x('C:addressbook-multiget', $Self->NS(),
+      x('D:prop',
+        x('D:getetag'),
+        x('C:address-data'),
+        map { x(join ":", @$_) } @$Props,
       ),
-      Depth => '0',
-    );
+      map { x('D:href', $_) } @$Urls,
+    ),
+    Depth => '0',
+  );
 
-    my $NS_C = $Self->ns('C');
-    my $NS_D = $Self->ns('D');
-    foreach my $Response (@{$Response->{"{$NS_D}response"} || []}) {
-      foreach my $Propstat (@{$Response->{"{$NS_D}propstat"} || []}) {
-        my $VCard = eval { $Self->_ParseReportData($Response, $Propstat, $Props) } || do {
-          push @Errors, $@ if $@;
-          next;
-        };
+  my $NS_C = $Self->ns('C');
+  my $NS_D = $Self->ns('D');
+  foreach my $Response (@{$Response->{"{$NS_D}response"} || []}) {
+    my $href = $Response->{"{$NS_D}href"}{content};
+    next unless $href;
+    foreach my $Propstat (@{$Response->{"{$NS_D}propstat"} || []}) {
+      my $etag = $Propstat->{"{$NS_D}prop"}{"{$NS_D}getetag"}{content} || '';
+      my $VCard = eval { $Self->_ParseReportData($Response, $Propstat, $Props) } || do {
+        push @Errors, $@ if $@;
+        next;
+      };
 
-        push @Contacts, $VCard;
-      }
+      push @Contacts, $VCard;
+
+      $ETags{$href} = $etag;
     }
   }
 
-  return wantarray ? (\@Contacts, \@Errors) : \@Contacts;
+  return wantarray ? (\@Contacts, \@Errors, \%ETags) : \@Contacts;
 }
 
 =head2 $self->SyncContacts($Path, $Props, %Args)
@@ -530,7 +565,33 @@ list context.
 
 sub SyncContacts {
   my ($Self, $Path, $Props, %Args) = @_;
-  $Props //= [];
+
+  my ($Links, $Removed, $Errors, $SyncToken) = $Self->SyncContactLinks($Path, %Args);
+
+  my @AllUrls = sort keys %$Links;
+
+  my ($Contacts, $ThisErrors, $ETags) = $Self->GetContactsMulti($Path, \@AllUrls, $Props);
+  push @$Errors, @$ThisErrors;
+
+  return wantarray ? ($Contacts, $Removed, $Errors, $SyncToken, $ETags) : $Contacts;
+}
+
+=head2 $self->SyncContactLinks($Path, %Args)
+
+uses the argument 'syncToken' to find newly added and removed
+cards from the server.
+
+Returns a list of:
+
+* Hash of href to etag for added/changed cargs
+* List of href of removed cards
+* List of errors
+* Scalar value of new syncToken
+
+=cut
+
+sub SyncContactLinks {
+  my ($Self, $Path, %Args) = @_;
 
   $Path || confess "Sync contacts path required";
 
@@ -543,25 +604,17 @@ sub SyncContacts {
       x('D:sync-level', 1),
       x('D:prop',
         x('D:getetag'),
-        x('C:address-data'),
-        map { x(join ":", @$_) } @$Props,
       ),
     ),
   );
 
-  if (($Response->{error} // "") eq 'valid-sync-token') {
-    delete $Args{syncToken};
-    return $Self->SyncContacts($Path, $Props, %Args);
-  }
-
-  my (@Contacts, @Removed, @Errors);
+  my (%Added, @Removed, @Errors);
 
   my $NS_C = $Self->ns('C');
   my $NS_D = $Self->ns('D');
   foreach my $Response (@{$Response->{"{$NS_D}response"} || []}) {
-    my $HRef = $Response->{"{$NS_D}href"}{content}
+    my $href = $Response->{"{$NS_D}href"}{content}
       || next;
-    my $CPath = $Self->_unrequest_url($HRef);
 
     # For members that have been removed, the DAV:response MUST
     # contain one DAV:status with a value set to '404 Not Found' and
@@ -569,10 +622,10 @@ sub SyncContacts {
     if (!$Response->{"{$NS_D}propstat"}) {
       my $Status = $Response->{"{$NS_D}status"}{content};
       if ($Status =~ m/ 404 /) {
-        push @Removed, $CPath;
+        push @Removed, $href;
       } else {
         warn "ODD STATUS";
-        push @Errors, "Odd status in non-propstat response: $Status";
+        push @Errors, "Odd status in non-propstat response $href: $Status";
       }
       next;
     }
@@ -585,12 +638,8 @@ sub SyncContacts {
       my $Status = $Propstat->{"{$NS_D}status"}{content};
 
       if ($Status =~ m/ 200 /) {
-        my $VCard = eval { $Self->_ParseReportData($Response, $Propstat, $Props) } || do {
-          push @Errors, $@ if $@;
-          next;
-        };
-
-        push @Contacts, $VCard;
+        my $etag = $Propstat->{"{$NS_D}prop"}{"{$NS_D}getetag"}{content};
+        $Added{$href} = $etag;
       }
       elsif ($Status =~ m/ 404 /) {
         # Missing properties return 404 status response, ignore
@@ -598,14 +647,14 @@ sub SyncContacts {
       }
       else {
         warn "ODD STATUS";
-        push @Errors, "Odd status in propstat response: $Status";
+        push @Errors, "Odd status in propstat response $href: $Status";
       }
     }
   }
 
   my $SyncToken = $Response->{"{$NS_D}sync-token"}{content};
 
-  return wantarray ? (\@Contacts, \@Removed, \@Errors, $SyncToken) : \@Contacts;
+  return (\%Added, \@Removed, \@Errors, $SyncToken);
 }
 
 =head2 $self->MoveContact($Path, $NewPath)
